"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Server = void 0;
const express_1 = __importDefault(require("express"));
const fs_1 = __importDefault(require("fs"));
const promises_1 = require("fs/promises");
const p5_analysis_1 = require("p5-analysis");
const path_1 = __importDefault(require("path"));
const pug_1 = __importDefault(require("pug"));
const stream_1 = require("stream");
const utils_1 = require("../utils");
const browserScriptEventRelay_1 = require("./browserScriptEventRelay");
const createDirectoryListing_1 = require("./createDirectoryListing");
const http_server_sync_1 = require("./http-server-sync");
const liveReload_1 = require("./liveReload");
const templates_1 = require("./templates");
const defaultServerOptions = {
    liveServer: true,
    logConsoleEvents: false,
    port: 3000,
    relayConsoleMessages: false,
    scanPorts: true,
    screenshot: null,
    theme: 'directory',
};
function createRouter(config) {
    const router = express_1.default.Router();
    router.get('/', async (req, res) => {
        const sketchFile = config.sketchFile;
        const file = path_1.default.join(config.root, decodeURIComponent(req.path));
        if (sketchFile) {
            if (await p5_analysis_1.Sketch.isSketchScriptFile(sketchFile)) {
                const sketch = await p5_analysis_1.Sketch.fromFile(sketchFile);
                sendHtml(req, res, await sketch.getHtmlContent());
            }
            else {
                sendHtml(req, res, await (0, promises_1.readFile)(sketchFile, 'utf-8'));
            }
        }
        else if (config.screenshot) {
            const { sketches } = fs_1.default.statSync(file).isDirectory()
                ? await p5_analysis_1.Sketch.analyzeDirectory(file, {
                    exclusions: createDirectoryListing_1.defaultDirectoryExclusions,
                })
                : { sketches: [] };
            if (sketches.length !== 1)
                throw new Error(`Expected exactly one sketch in ${file}`);
            const [sketch] = sketches;
            const html = sketch.htmlFile
                ? await (0, promises_1.readFile)(sketch.htmlFilePath, 'utf-8')
                : await sketch.getHtmlContent();
            sendHtml(req, res, html);
        }
        else {
            await sendDirectoryListing(config, req, res);
        }
    });
    router.post('/__p5_server/screenshot', express_1.default.json({ limit: '50mb' }), async (req, res) => {
        const { dataURL } = req.body;
        const m = dataURL.match(/^data:image\/(.+?);base64,(.*)$/);
        if (!m || !config.screenshot?.onFrameData)
            return res.sendStatus(200);
        await config.screenshot.onFrameData({
            imageType: m[1],
            data: Buffer.from(m[2], 'base64'),
            frameNumber: req.body.frameNumber,
        });
        res.sendStatus(200);
    });
    router.get('/*.html?', (req, res, next) => {
        const file = path_1.default.join(config.root, decodeURIComponent(req.path));
        try {
            if (req.query.fmt === 'view') {
                res.set('Content-Type', 'text/plain');
                res.sendFile(req.path, { root: config.root });
                return;
            }
            if (req.headers['accept']?.match(/\btext\/html\b/)) {
                sendHtml(req, res, fs_1.default.readFileSync(file, 'utf-8'));
                return;
            }
        }
        catch (e) {
            if (e.code !== 'ENOENT') {
                throw e;
            }
        }
        next();
    });
    router.get('/*.js', async (req, res, next) => {
        const filepath = path_1.default.join(config.root, decodeURIComponent(req.path));
        if (req.headers['accept']?.match(/\btext\/html\b/) &&
            req.query.fmt !== 'view' &&
            (await p5_analysis_1.Sketch.isSketchScriptFile(filepath))) {
            const { sketches } = await p5_analysis_1.Sketch.analyzeDirectory(path_1.default.dirname(filepath), {
                exclusions: createDirectoryListing_1.defaultDirectoryExclusions,
            });
            const sketch = sketches.find(sketch => sketch.files.includes(path_1.default.basename(filepath)));
            if (sketch) {
                return sendHtml(req, res, await sketch.getHtmlContent());
            }
        }
        if (req.headers['accept']?.match(/\btext\/html\b/) && req.query.fmt === 'view') {
            const source = await (0, promises_1.readFile)(filepath, 'utf-8');
            const title = req.path.replace(/^\//, '');
            const html = (0, templates_1.sourceViewTemplate)({ source, title });
            res.set('Content-Type', 'text/html');
            return res.send(html);
        }
        try {
            const errs = p5_analysis_1.Script.fromFile(filepath).getErrors();
            if (errs.length) {
                res.set('Content-Type', 'text/html');
                return res.send((0, templates_1.createSyntaxErrorJsReporter)(errs, filepath));
            }
        }
        catch (e) {
            if (e.code === 'ENOENT') {
                return next();
            }
            else {
                throw e;
            }
        }
        next();
    });
    router.get('/*.md', (req, res, next) => {
        if (req.headers['accept']?.match(/\btext\/html\b/)) {
            const file = path_1.default.join(config.root, decodeURIComponent(req.path));
            if (!fs_1.default.existsSync(file)) {
                return next();
            }
            const data = fs_1.default.readFileSync(file, 'utf-8');
            res.set('Content-Type', 'text/html');
            return res.send((0, templates_1.markdownToHtmlPage)(data));
        }
        return next();
    });
    router.get('*', (req, res, next) => {
        if (req.headers['accept']?.match(/\btext\/html\b/)) {
            const file = path_1.default.join(config.root, decodeURIComponent(req.path));
            if (fs_1.default.existsSync(file) && fs_1.default.statSync(file).isDirectory()) {
                return sendDirectoryListing(config, req, res);
            }
        }
        next();
    });
    return router;
    function sendHtml(req, res, html) {
        html = (0, liveReload_1.injectLiveReloadScript)(html, req.app.locals.liveReloadServer);
        if (config.relayConsoleMessages ||
            'send-console-messages' in req.query ||
            'vscodeBrowserReqId' in req.query) {
            html = (0, browserScriptEventRelay_1.injectScriptEventRelayScript)(html);
        }
        if (config.screenshot && req.path === '/') {
            html = (0, utils_1.addScriptToHtmlHead)(html, '/__p5_server_static/screenshot.min.js');
            html = (0, utils_1.addScriptToHtmlHead)(html, {
                __p5_server_screenshot_settings: config.screenshot,
            });
        }
        res.set('Content-Type', 'text/html');
        res.send(html);
    }
}
async function sendDirectoryListing(config, req, res) {
    if (!req.originalUrl.endsWith('/')) {
        return res.redirect(req.originalUrl + '/');
    }
    const dir = path_1.default.join(config.root, decodeURIComponent(req.path).replace(/\//g, path_1.default.sep));
    const indexFile = (await (0, promises_1.readdir)(dir)).find(file => /^index\.html?$/i.test(file));
    let html = indexFile
        ? await (0, promises_1.readFile)(path_1.default.join(dir, indexFile), 'utf-8')
        : await (0, createDirectoryListing_1.createDirectoryListing)(dir, req.originalUrl, {
            templateName: config.theme,
        });
    if (config.liveServer) {
        html = (0, liveReload_1.injectLiveReloadScript)(html, req.app.locals.liveReloadServer);
    }
    return res.send(html);
}
async function startServer(config, sketchRelay) {
    const mountPoints = config.mountPoints;
    const app = (0, express_1.default)();
    app.use('/__p5_server_static', express_1.default.static(path_1.default.join(__dirname, 'static')));
    for (const { filePath, urlPath } of mountPoints) {
        let root = filePath;
        let sketchFile;
        if (!fs_1.default.statSync(root).isDirectory()) {
            sketchFile = root;
            root = path_1.default.dirname(root);
        }
        const routerConfig = { ...config, root, sketchFile };
        app.use(urlPath, createRouter(routerConfig));
        app.use(urlPath, express_1.default.static(root));
    }
    if (mountPoints.length > 0) {
        const mountListTmpl = pug_1.default.compileFile(path_1.default.join(templates_1.templateDir, 'mountPoints.pug'));
        app.get('/', (_req, res) => res.send(mountListTmpl({ mountPoints })));
    }
    if (fs_1.default.statSync(mountPoints[0].filePath).isDirectory()) {
        (0, createDirectoryListing_1.createDirectoryListing)(mountPoints[0].filePath, mountPoints[0].urlPath, {
            templateName: config.theme,
        });
    }
    let server = null;
    const port = config.port;
    for (let p = port; p < port + 10; p++) {
        try {
            server = await (0, http_server_sync_1.listenSync)(app, p);
            break;
        }
        catch (e) {
            if (e.code !== 'EADDRINUSE' || !config.scanPorts) {
                throw e;
            }
            console.log(`Port ${p} is in use, retrying...`);
        }
    }
    if (!server)
        server = await (0, http_server_sync_1.listenSync)(app);
    const address = server.address();
    if (!address || typeof address === 'string') {
        throw new Error('Failed to start the server');
    }
    (0, browserScriptEventRelay_1.attachBrowserScriptRelay)(server, sketchRelay);
    try {
        const liveReloadServer = await (0, liveReload_1.createLiveReloadServer)({
            port: Math.min(port + 35729 - config.port, 30000),
            scanPorts: true,
            watchDirs: [templates_1.templateDir, ...mountPoints.map(mount => mount.filePath)],
        });
        app.locals.liveReloadServer = liveReloadServer;
        const url = `http://localhost:${address.port}`;
        return { server, liveReloadServer, url };
    }
    catch (e) {
        server.close();
        throw e;
    }
}
class Server {
    constructor(options = {}) {
        var _a;
        this.server = null;
        this.liveReloadServer = null;
        this.browserScriptEmitter = new stream_1.EventEmitter();
        this.emitScriptEvent = this.browserScriptEmitter.emit.bind(this.browserScriptEmitter);
        this.onScriptEvent = this.browserScriptEmitter.on.bind(this.browserScriptEmitter);
        const mountPoints = options.mountPoints && options.mountPoints.length > 0
            ? Server.normalizeMountPoints(options.mountPoints)
            : [{ filePath: options.root || '.', urlPath: '/' }];
        this.mountPoints = mountPoints;
        this.config = { ...defaultServerOptions, root: null, ...options, mountPoints };
        (_a = this.config).theme || (_a.theme = defaultServerOptions.theme);
    }
    static async start(options = {}) {
        return new Server(options).start();
    }
    async start() {
        const { server, liveReloadServer, url } = await startServer(this.config, this);
        this.server = server;
        this.liveReloadServer = liveReloadServer;
        this.url = url;
        return this;
    }
    async stop() {
        if (this.server) {
            await (0, http_server_sync_1.closeSync)(this.server);
            this.server = null;
        }
        this.url = undefined;
        this.liveReloadServer?.close();
        this.liveReloadServer = null;
    }
    filePathToUrl(filePath) {
        const baseUrl = this.url || `http://localhost:${this.config.port}`;
        for (const mountPoint of this.mountPoints) {
            const filePrefix = mountPoint.filePath + path_1.default.sep;
            const pathPrefix = mountPoint.urlPath.replace(/(?<!\/)$/, '/');
            if (filePath.startsWith(filePrefix)) {
                return baseUrl + filePath.replace(filePrefix, pathPrefix);
            }
        }
        return null;
    }
    urlPathToFilePath(urlPath) {
        for (const mountPoint of this.mountPoints) {
            const filePrefix = mountPoint.filePath + path_1.default.sep;
            const pathPrefix = mountPoint.urlPath.replace(/(?<!\/)$/, '/');
            if (urlPath.startsWith(pathPrefix)) {
                return urlPath.replace(pathPrefix, filePrefix);
            }
        }
        return null;
    }
    static normalizeMountPoints(mountPoints) {
        const finalPathSep = new RegExp(`${path_1.default.sep}$`);
        const mounts = mountPoints
            .map(mount => (typeof mount === 'string' ? { filePath: mount } : mount))
            .map(mount => ({
            urlPath: '/' + (mount.name || path_1.default.basename(mount.filePath)),
            ...mount,
        }))
            .map(mount => ({ ...mount, urlPath: mount.urlPath.replace(/ /g, ' ') }))
            .map(mount => ({ ...mount, filePath: mount.filePath.replace(/\//g, path_1.default.sep) }))
            .map(mount => ({
            ...mount,
            filePath: mount.filePath.replace(finalPathSep, ''),
            urlPath: mount.urlPath.replace(/\/$/, ''),
        }));
        const seen = new Set();
        for (const mount of mounts) {
            if (seen.has(mount.urlPath)) {
                mount.urlPath = findUniqueName(mount.urlPath, seen);
            }
            seen.add(mount.urlPath);
        }
        return mounts;
        function findUniqueName(base, exclude) {
            for (const name of generateNames(base)) {
                if (!exclude.has(name)) {
                    return name;
                }
            }
            return null;
        }
        function* generateNames(base) {
            yield base;
            let ix = 2;
            const m = base.match(/^(.*?)-(\d*)$/);
            if (m) {
                base = m[1];
                ix = parseInt(m[2], 10) + 1;
            }
            while (true) {
                yield `${base}-${ix++}`;
            }
        }
    }
}
exports.Server = Server;
